
"""
Simple model of a vector-borne disease with 10% host mortality spreading among
hosts and vectors that have natural birth and death rates in a single
population. There is no evolution and pathogen genomes don't affect spread.
"""

from opqua.model import Model # This is all you need to import to run models.

my_model = Model() # Make a new model object.
my_model.newSetup('my_setup', preset='vector-borne',
        # Create a new set of parameters called "my_setup" to be used to
        # simulate a population in the model. Use the default parameter set for
        # a vector-borne model.
    mortality_rate_host=1e-2,
        # change the default host mortality rate to 10% of recovery rate
    protection_upon_recovery_host=[0,10],
        # make hosts immune to the genome that infected them if they recover
        # [0,10] means that pathogen genome positions 0 through 9 will be saved
        # as immune memory
    birth_rate_host=1.5e-2,
        # change the default host birth rate to 0.015 births/time unit
    death_rate_host=1e-2,
        # change the default natural host death rate to 0.01 births/time unit
    birth_rate_vector=1e-2,
        # change the default vector birth rate to 0.01 births/time unit
    death_rate_vector=1e-2
        # change the default natural vector death rate to 0.01 deaths/time unit
    )

my_model.newPopulation(
    'my_population', 'my_setup', num_hosts=100, num_vectors=100
    )
    # Create a new population of 100 hosts and 100 vectors called
    # "my_population". The population uses parameters stored in "my_setup".
my_model.addPathogensToHosts( 'my_population',{'AAAAAAAAAA':20} )
    # Add pathogens with a genome of "AAAAAAAAAA" to 20 random hosts in
    # population "my_population".

my_model.run(0,200) # Run the simulation for 200 time units.
data = my_model.saveToDataFrame('vector-borne_birth-death_example.csv')
    # Save the model results to a table.
graph = my_model.compartmentPlot('vector-borne_birth-death_example.png', data)
    # Plot the number of susceptible and infected hosts in the model over time.




"""
Simple host-host model with susceptible and infected hosts in a single
population scenario with no evolution.
"""

from opqua.model import Model # This is all you need to import to run models.

my_model = Model() # Make a new model object.
my_model.newSetup('my_setup', preset='host-host')
    # Create a new set of parameters called "my_setup" to be used to simulate
    # a population in the model. Use the default parameter set for a
    # host-host transmission model.
my_model.newPopulation('my_population', 'my_setup', num_hosts=100)
    # Create a new population of 100 hosts and 0 vectors called
    # "my_population". The population uses parameters stored in "my_setup".
my_model.addPathogensToHosts( 'my_population',{'AAAAAAAAAA':20} )
    # Add pathogens with a genome of "AAAAAAAAAA" to 20 random hosts in
    # population "my_population".
my_model.run(0,200) # Run the simulation for 200 time units.
data = my_model.saveToDataFrame('Basic_example.csv')
    # Save the model results to a table.
graph = my_model.compartmentPlot('Basic_example.png', data)
    # Plot the number of susceptible and infected hosts in the model over time.




"""
Vector-borne model with susceptible and infected hosts and vectors, showing a
metapopulation model setup with multiple populations connected to each other by
"population contact" events between vectors and hosts, in which a vector and a
host from different populations contact each other without migrating from one
population to another.

Population A is connected to Population B and to Clustered Population 4 (both
are one-way connections).
Clustered Populations 0-4 are all connected to each other in two-way
connections.
Isolated population is not connected to any others.

Two different pathogen genotypes are initially seeded into Populations A and B.
"""

from opqua.model import Model

model = Model()
model.newSetup('setup_normal', preset='vector-borne')
model.newSetup(
    'setup_cluster',
    contact_rate_host_vector = ( 2 *
        model.setups['setup_normal'].contact_rate_host_vector ),
    preset='vector-borne'
    ) # uses default parameters but doubles contact rate of the first setup

model.newPopulation('population_A','setup_normal', num_hosts=20, num_vectors=20)
model.newPopulation('population_B','setup_normal', num_hosts=20, num_vectors=20)
    # Create two populations that will be connected.
model.newPopulation(
    'isolated_population','setup_normal', num_hosts=20, num_vectors=20
    ) # A third population will remain isolated.

model.createInterconnectedPopulations(
    5,'clustered_population_','setup_cluster',
    host_migration_rate=0, vector_migration_rate=0,
    vector_host_contact_rate=2e-2,
    host_vector_contact_rate=2e-2,
    num_hosts=20, num_vectors=20
    )
    # Create a cluster of 5 populations connected to each other with a
    # population contact rate of 1e-2 between each of them in both directions.
    # Each population has an numbered ID with the prefix
    # "clustered_population_", has the parameters defined in the "setup_cluster"
    # setup, and has 20 hosts and vectors.
model.linkPopulationsHostVectorContact('population_A','clustered_population_4',2e-2)
model.linkPopulationsVectorHostContact('clustered_population_4','population_A',2e-2)
    # We link population_A to one of the clustered populations with a one-way
    # population contact rate of 1e-2 for population_A hosts and
    # clustered_population_4 vectors. Note that for population contacts, both
    # populations need to have contact rates towards each other (migration does
    # not require this)
model.linkPopulationsHostVectorContact('population_A','population_B',2e-2)
model.linkPopulationsVectorHostContact('population_B','population_A',2e-2)
    # We link population_A to population_B with a one-way
    # population contact rate of 1e-2 for population_A hosts and
    # population_B vectors. Note that for population contacts, both
    # populations need to have contact rates towards each other (migration does
    # not require this)

model.addPathogensToHosts( 'population_A',{'AAAAAAAAAA':5} )
    # population_A starts with AAAAAAAAAA genotype pathogens.
model.addPathogensToHosts( 'population_B',{'GGGGGGGGGG':5} )
    # population_B starts with GGGGGGGGGG genotype pathogens.

output = model.run(0,100,time_sampling=0)
data = model.saveToDataFrame('metapopulations_population_contact_example.csv')
graph = model.populationsPlot( # Plot infected hosts per population over time.
    'metapopulations_population_contact_example.png', data,
    num_top_populations=8, track_specific_populations=['isolated_population'],
        # Make sure to plot th isolated population totals if not in the top
        # infected populations.
    y_label='Infected hosts' # change y label
    )




"""
Host-host transmission model with susceptible and infected hosts in a single
population scenario, illustrating pathogen evolution through independent
reassortment/segregation of chromosomes, increased transmissibility,
and intra-host competition.

When two pathogens with different genomes meet in the same host (or vector),
the pathogen with the most fit genome has a higher probability of being
transmitted to another host (or vector). In this case, the transmission rate
DOES vary according to genome, with more fit genomes having a higher
transmission rate. Once an event occurs, the pathogen with higher fitness also
has a higher likelihood of being transmitted.

Here, we define a landscape of stabilizing selection where there is an optimal
genome and every other genome is less fit, but fitness functions can be defined
in any arbitrary way (accounting for multiple peaks, for instance, or special
cases for a specific genome sequence).
"""

from opqua.model import Model

my_optimal_genome = 'BEST/BEST/BEST/BEST'
    # Define an optimal genome. "/" denotes separators between different
    # chromosomes, which are segregated and recombined independently of each
    # other (this model has no recombination).

# Define custom fitness function for the host:
# Fitness functions must take in 1 argument and return a positive number as a
# fitness value. Here, we take advantage of one of the preset functions, but you
# can define it any way you want!
def myHostFitness(genome):
    return Model.peakLandscape(
        genome, peak_genome=my_optimal_genome, min_value=1e-10
        )
        # Stabilizing selection: any deviation from the "optimal genome"
        # sequence results in an exponential decay in fitness to the min_fitness
        # value at the maximum possible distance. Here we use strong selection,
        # with a very low minimum fitness.

def myHostContact(genome):
    return 1 if genome == my_optimal_genome else 0.05
        # Stabilizing selection: any deviation from the "optimal genome"
        # sequence 1/20 of the fitness of the optimal genome. There is no
        # middle ground between the optimal and the rest, in this case.

model = Model()
model.newSetup( # Now, we'll define our new setup:
    'my_setup', preset='host-host', # Use default host-host parameters.
    possible_alleles='ABDEST',
        # Define "letters" in the "genome", or possible alleles for each locus.
        # Each locus can have different possible alleles if you define this
        # argument as a list of strings, but here, we take the simplest
        # approach.
    num_loci=len(my_optimal_genome),
        # Define length of "genome", or total number of alleles.
    contact_rate_host_host = 2e0,
    contactHost=myHostContact,
        # Assign the contact function we created (could be a lambda function)
    fitnessHost=myHostFitness,
        # Assign the fitness function we created (could be a lambda function)
    recombine_in_host=1e-3,
        # Modify "recombination" rate of pathogens when in host to get some
        # evolution! This can either be independent segregation of chromosomes
        # (equivalent to reassortment), recombination of homologous chromosomes,
        # or a combination of both.
    num_crossover_host=0
        # By specifying the average number of crossover events that happen
        # during recombination to be zero, we ensure that "recombination" is
        # restricted to independent segregation of chromosomes (separated by
        # "/").
    )

model.newPopulation('my_population','my_setup',num_hosts=100)
model.addPathogensToHosts( 'my_population',{'BEST/BADD/BEST/BADD':10} )
    # We will start off the simulation with a suboptimal pathogen genome,
    # Throughout the course of the simulation, we should see this genome
    # be outcompeted by more optimal pathogen genotypes, culminating in the
    # optimal genome, which outcompetes all others.
model.addPathogensToHosts( 'my_population',{'BADD/BEST/BADD/BEST':10} )
    # We will start off the simulation with a second suboptimal pathogen genome,
    # Throughout the course of the simulation, we should see this genome
    # be outcompeted by more optimal pathogen genotypes, culminating in the
    # optimal genome, which outcompetes all others.
model.run(0,500,time_sampling=100)
data = model.saveToDataFrame('transmissibility_function_reassortment_example.csv')

graph_composition = model.compositionPlot(
        # Create a plot to track pathogen genotypes across time.
    'transmissibility_function_reassortment_example_composition.png', data
    )

graph_clustermap = model.clustermap(
    'transmissibility_function_reassortment_example_clustermap.png', data,
    save_data_to_file='transmissibility_function_reassortment_example_pairwise_distances.csv',
    num_top_sequences=24
    )
    # Generate a heatmap and dendrogram for the top 24 genomes. Besides creating
    # the plot, outputs the pairwise distance matrix to a csv file as well.

graph_compartments = model.compartmentPlot(
    'transmissibility_function_reassortment_example_compartments.png', data
    )
    # Also generate a normal compartment plot. Notice the total number of
    # infections in the composition plot can exceed the number of infected hosts
    # in the compartment plot. This happens because a single host infected by
    # multiple genotypes is counted twice in the former, but not the latter.




"""
Host-host transmission model with susceptible and infected hosts in a single
population scenario, illustrating pathogen evolution through de novo mutations
and intra-host competition.

When two pathogens with different genomes meet in the same host (or vector),
the pathogen with the most fit genome has a higher probability of being
transmitted to another host (or vector). In this case, the transmission rate
does NOT vary according to genome. Once an event occurs, however, the pathogen
with higher fitness has a higher likelihood of being transmitted.

Here, we define a landscape of stabilizing selection where there is an optimal
genome and every other genome is less fit, but fitness functions can be defined
in any arbitrary way (accounting for multiple peaks, for instance, or special
cases for a specific genome sequence).
"""

from opqua.model import Model

my_optimal_genome = 'BEST' # Define an optimal genome.

# Define custom fitness function for the host:
# Fitness functions must take in 1 argument and return a positive number as a
# fitness value. Here, we take advantage of one of the preset functions, but you
# can define it any way you want!
def myHostFitness(genome):
    return Model.peakLandscape(
        genome, peak_genome=my_optimal_genome, min_value=1e-10
        )
        # Stabilizing selection: any deviation from the "optimal genome"
        # sequence results in an exponential decay in fitness to the min_fitness
        # value at the maximum possible distance. Here we use strong selection,
        # with a very low minimum fitness.

model = Model()
model.newSetup( # Now, we'll define our new setup:
    'my_setup', preset='host-host', # Use default host-host parameters.
    possible_alleles='ABDEST',
        # Define "letters" in the "genome", or possible alleles for each locus.
        # Each locus can have different possible alleles if you define this
        # argument as a list of strings, but here, we take the simplest
        # approach.
    num_loci=len(my_optimal_genome),
        # Define length of "genome", or total number of alleles.
    fitnessHost=myHostFitness,
        # Assign the fitness function we created (could be a lambda function)
    mutate_in_host=5e-2
        # Modify de novo mutation rate of pathogens when in host to get some
        # evolution!
    )

model.newPopulation('my_population','my_setup', num_hosts=100)
model.addPathogensToHosts( 'my_population',{'BADD':10} )
    # We will start off the simulation with a suboptimal pathogen genome,
    # "BADD". Throughout the course of the simulation, we should see this genome
    # be outcompeted by more optimal pathogen genotypes, culminating in the
    # optimal genome, "BEST", which outcompetes all others.
model.run(0,200)
data = model.saveToDataFrame(
    'fitness_function_mutation_example.csv'
    )

graph_composition = model.compositionPlot(
        # Create a plot to track pathogen genotypes across time.
    'fitness_function_mutation_example_composition.png', data,
    num_top_sequences=6,
        # Track the 6 most represented genomes overall (remaining genotypes are
        # lumped into the "Other" category).
    track_specific_sequences=['BADD']
        # Include the initial genome in the graph if it isn't in the top 6.
    )

graph_clustermap = model.clustermap(
    'fitness_function_mutation_example_clustermap.png', data,
    save_data_to_file='fitness_function_mutation_example_pairwise_distances.csv',
    num_top_sequences=15,
    track_specific_sequences=['BADD']
    )
    # Generate a heatmap and dendrogram for the top 15 genomes, include the
    # ancestral genome "BADD" in the phylogeny. Besides creating the plot,
    # outputs the pairwise distance matrix to a csv file as well.

graph_compartments = model.compartmentPlot(
    'fitness_function_example_reassortment_compartments.png', data
    )
    # Also generate a normal compartment plot. Notice the total number of
    # infections in the composition plot can exceed the number of infected hosts
    # in the compartment plot. This happens because a single host infected by
    # multiple genotypes is counted twice in the former, but not the latter.




"""
Vector-borne model with susceptible and infected hosts and vectors in a single
population scenario, showing the effect of various interventions done at
different time points.

For more information on how each intervention function works, check out the
documentation for each function fed into newIntervention() in the README file.
"""

from opqua.model import Model

model = Model()
model.newSetup('my_setup',preset='vector-borne')
model.newSetup(
    'my_setup_2', contact_rate_host_vector=4e-1, preset='vector-borne'
    )
    # We make a second setup with the same parameters, but duplicate the contact
    # rate.

model.newPopulation('my_population','my_setup', num_hosts=100, num_vectors=100)
model.addPathogensToHosts( 'my_population',{'AAAAAAAAAA':20} )

model.newIntervention(
    20, 'addPathogensToHosts',
    [ 'my_population', {'TTTTTTTTTT':5, 'CCCCCCCCCC':5, } ]
    )
    # At time 20, adds pathogens of genomes TTTTTTTTTT and CCCCCCCCCC to 5
    # random hosts each.

model.newIntervention( 50, 'addVectors', [ 'my_population', 10 ] )
    # At time 50, adds 10 healthy vectors to population.
model.newIntervention(
    50, 'newVectorGroup',
    [ 'my_population', '10_new_vectors', 10, 'healthy' ]
    )
    # At time 50, selects 10 healthy vectors from population 'my_population'
    # and stores them under the group ID '10_new_vectors'.
model.newIntervention(
    50, 'addPathogensToVectors',
    [ 'my_population', {'GGGGGGGGGG':10}, '10_new_vectors' ]
    )
    # At time 50, adds pathogens of genomes GGGGGGGGGG to 10
    # random hosts in the '10_new_vectors' group (so, all 10 of them).
    # The last '10_new_vectors' argument specifies which group to sample
    # from (if not specified, sampling occurs from whole population).

model.newIntervention( 100, 'setSetup', [ 'my_population', 'my_setup_2' ] )
    # At time 100, changes the parameters of my_population to those in
    # my_setup_2, with twice the contact rate.

model.newIntervention(
    150, 'newHostGroup',
    [ 'my_population', 'treated_hosts', -1, 'infected' ]
    )
    # At time 150, selects 100% of infected hosts and stores them under the
    # group ID 'treated_hosts'. The third argument selects all hosts available
    # when set to -1, as above.
model.newIntervention(
    150, 'newVectorGroup',
    [ 'my_population', 'treated_vectors', -1, 'infected' ]
    )
    # At time 150, selects 100% of infected vectors and stores them under the
    # group ID 'treated_vectors'. The third argument selects all vectors
    # available when set to -1, as above.
model.newIntervention(
    150, 'treatHosts',
    [ 'my_population', 1, 'GGGGGGGGGG', 'treated_hosts' ]
    )
    # At time 150, treat 100% of the "treated_hosts" population with a treatment
    # that kills pathogens unless they contain a 'GGGGGGGGGG' sequence in their
    # genome.
model.newIntervention(
    150, 'treatVectors',
    [ 'my_population', 1, 'GGGGGGGGGG', 'treated_vectors' ]
    )
    # At time 150, treat 100% of the "treated_vectors" population with a
    # treatment that kills pathogens unless they contain a 'GGGGGGGGGG' sequence
    # in their genome.

model.newIntervention(
    250, 'newHostGroup',
    [ 'my_population', 'vaccinated', 0.85, 'any' ]
    )
    # At time 250, selects 85% of random hosts and stores them under the group
    # ID 'vaccinated'. They may be healthy or infected.
model.newIntervention(
    250, 'protectHosts',
    [ 'my_population', 1, 'GGGGGGGGGG', 'vaccinated' ]
    )
# model.newIntervention( 250, model.protectHosts, [ 'my_population', 1, 'C', 'vaccinated' ] )
# model.newIntervention( 250, model.protectHosts, [ 'my_population', 1, 'T', 'vaccinated' ] )
    # At time 250, protects 100% of the vaccinated group from pathogens
    # with a 'GGGGGGGGGG' sequence in their genome.

output = model.run(0,400)
data = model.saveToDataFrame('intervention_examples.csv')
graph = model.compositionPlot( 'intervention_examples_composition.png', data )
    # Create a plot to track pathogen genotypes across time.
graph = model.compartmentPlot('intervention_examples_compartments.png', data)
    # Also generate a normal compartment plot. Notice the total number of
    # infections in the composition plot can exceed the number of infected hosts
    # in the compartment plot. This happens because a single host infected by
    # multiple genotypes is counted twice in the former, but not the latter.
